import unittest
from pathlib import Path
from unittest.mock import Mock, mock_open, patch

import typer
from pydantic import ValidationError
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from jupyter_deploy.constants import MANIFEST_FILENAME
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.base_project_handler import (
    BaseProjectHandler,
    NotADictError,
    retrieve_project_manifest,
    retrieve_project_manifest_if_available,
    retrieve_variables_config,
)


class TestBaseProjectHandler(unittest.TestCase):
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_calls_retrieve_project_and_save_attributes(self, mock_cwd: Mock, mock_retrieve: Mock) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = EngineType.TERRAFORM
        mock_retrieve.return_value = mock_manifest

        # Execute
        handler = BaseProjectHandler()

        # Assert
        mock_retrieve.assert_called_once_with(Path("/fake/path/manifest.yaml"))
        self.assertEqual(handler.engine, EngineType.TERRAFORM)
        self.assertEqual(handler.project_manifest, mock_manifest)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_print_on_filenotfound_error(
        self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = FileNotFoundError("Missing jupyter-deploy manifest.")

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        # Verify console output
        mock_console.print.assert_any_call(
            ":x: The path does not correspond to a jupyter-deploy project.", style="bold red"
        )
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call("Reason: could not find the jupyter-deploy manifest file.", style="red")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_print_on_os_error(self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = OSError("Permission denied")

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        # Verify console output
        mock_console.print.assert_any_call(":x: Could not access the jupyter-deploy manifest.", style="bold red")
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call(
            "Reason: OS error when reading the jupyter-deploy manifest file.", style="red"
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_prints_on_runtime_error(
        self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = NotADictError("Invalid type: jupyter-deploy manifest is not a dict.")

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        # Verify console output
        mock_console.print.assert_any_call(":x: The jupyter-deploy manifest is invalid.", style="bold red")
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call(
            "Reason: expected the jupyter-deploy manifest file to parse as dict.", style="red"
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_prints_on_yamlparse_error(
        self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = ParserError("YAML parsing error")

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        # Verify console output
        mock_console.print.assert_any_call(":x: The jupyter-deploy manifest is invalid.", style="bold red")
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call(
            "Reason: cannot parse the jupyter-deploy manifest content as YAML.", style="red"
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_prints_on_yamlscanner_error(
        self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = ScannerError("YAML scanner error")

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        # Verify console output
        mock_console.print.assert_any_call(":x: The jupyter-deploy manifest is invalid.", style="bold red")
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call(
            "Reason: cannot parse the jupyter-deploy manifest content as YAML.", style="red"
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    @patch("rich.console.Console")
    def test_exits_and_prints_on_pydantic_error(
        self, mock_console_class: Mock, mock_cwd: Mock, mock_retrieve: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/fake/path")
        mock_console = Mock()
        mock_console_class.return_value = mock_console
        mock_retrieve.side_effect = ValidationError("some error", [])

        # Execute and Assert
        with self.assertRaises(typer.Exit):
            BaseProjectHandler()

        print(mock_console.print.mock_calls[1])

        # Verify console output
        mock_console.print.assert_any_call(":x: The jupyter-deploy manifest is invalid.", style="bold red")
        mock_console.line.assert_any_call()
        mock_console.print.assert_any_call(
            "Reason: the manifest file does not conform to the expected schema.", style="red"
        )


class TestRetrieveProjectManifestIfAvailable(unittest.TestCase):
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_returns_manifest_when_successful(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/fake/path")
        mock_manifest = Mock()
        mock_retrieve.return_value = mock_manifest

        # Execute
        result = retrieve_project_manifest_if_available(project_path)

        # Assert
        mock_retrieve.assert_called_once_with(project_path / MANIFEST_FILENAME)
        self.assertEqual(result, mock_manifest)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_returns_none_when_filenotfound_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/fake/path")
        mock_retrieve.side_effect = FileNotFoundError("Missing jupyter-deploy manifest.")

        # Execute
        result = retrieve_project_manifest_if_available(project_path)

        # Assert
        mock_retrieve.assert_called_once_with(project_path / MANIFEST_FILENAME)
        self.assertIsNone(result)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_returns_none_when_notadict_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/fake/path")
        mock_retrieve.side_effect = NotADictError("Invalid manifest: jupyter-deploy manifest is not a dict.")

        # Execute
        result = retrieve_project_manifest_if_available(project_path)

        # Assert
        mock_retrieve.assert_called_once_with(project_path / MANIFEST_FILENAME)
        self.assertIsNone(result)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_returns_none_when_validation_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/fake/path")
        mock_retrieve.side_effect = ValidationError("Invalid manifest schema", [])

        # Execute
        result = retrieve_project_manifest_if_available(project_path)

        # Assert
        mock_retrieve.assert_called_once_with(project_path / MANIFEST_FILENAME)
        self.assertIsNone(result)


class TestRetrieveProjectManifest(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.file_exists")
    def test_checks_file_existence(self, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = False
        manifest_path = Path("/fake/path/manifest.yaml")

        # Execute and Assert
        with self.assertRaises(FileNotFoundError):
            retrieve_project_manifest(manifest_path)

        mock_file_exists.assert_called_once_with(manifest_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="schema_version: 1\ntemplate:\n  name: test\n  engine: terraform\n  version: 1.0.0",
    )
    @patch("yaml.safe_load")
    @patch("jupyter_deploy.manifest.JupyterDeployManifest")
    def test_open_file_call_safe_load_and_parse(
        self, mock_manifest_class: Mock, mock_yaml_load: Mock, mock_open_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")
        yaml_content = {"schema_version": 1, "template": {"name": "test", "engine": "terraform", "version": "1.0.0"}}
        mock_yaml_load.return_value = yaml_content
        mock_manifest = Mock()
        mock_manifest_class.return_value = mock_manifest

        # Execute
        result = retrieve_project_manifest(manifest_path)

        # Assert
        mock_open_file.assert_called_once_with(manifest_path)
        mock_yaml_load.assert_called_once()
        mock_manifest_class.assert_called_once_with(**yaml_content)
        self.assertEqual(result, mock_manifest)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open)
    def test_parse_manifest_versions(self, mock_open_file: Mock, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")

        # Test for schema_version 1
        yaml_content = """
        schema_version: 1
        template:
          name: test
          engine: terraform
          version: 1.0.0
        """
        mock_open_file.return_value.read.return_value = yaml_content

        with patch(
            "yaml.safe_load",
            return_value={"schema_version": 1, "template": {"name": "test", "engine": "terraform", "version": "1.0.0"}},
        ):
            # Execute
            result = retrieve_project_manifest(manifest_path)

            # Assert
            self.assertEqual(result.template.name, "test")
            self.assertEqual(result.template.engine, EngineType.TERRAFORM)
            self.assertEqual(result.template.version, "1.0.0")

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open")
    def test_surfaces_error_when_open_raises_os_error(self, mock_open_file: Mock, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")
        mock_open_file.side_effect = OSError("Permission denied")

        # Execute and Assert
        with self.assertRaises(OSError):
            retrieve_project_manifest(manifest_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open, read_data="invalid: yaml: content:")
    @patch("yaml.safe_load")
    def test_raise_yaml_parse_error_on_invalid_yaml(
        self, mock_yaml_load: Mock, mock_open_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")
        mock_yaml_load.side_effect = ParserError("YAML parsing error")

        # Execute and Assert
        with self.assertRaises(ParserError):
            retrieve_project_manifest(manifest_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open, read_data="- item1\n- item2")
    @patch("yaml.safe_load")
    def test_raise_value_error_when_parsed_content_is_not_a_dict(
        self, mock_yaml_load: Mock, mock_open_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")
        mock_yaml_load.return_value = ["item1", "item2"]  # Not a dict

        # Execute and Assert
        with self.assertRaises(NotADictError) as context:
            retrieve_project_manifest(manifest_path)
        self.assertIn("Invalid manifest: jupyter-deploy manifest is not a dict.", str(context.exception))

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="schema_version: 1\ntemplate:\n  name: test\n  version: 1.0.0",
    )
    @patch("yaml.safe_load")
    def test_raise_validation_error_when_pydantic_parsing_fails(
        self, mock_yaml_load: Mock, mock_open_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        manifest_path = Path("/fake/path/manifest.yaml")
        # Missing required 'engine' field
        mock_yaml_load.return_value = {
            "schema_version": 1,
            "template": {
                "name": "test",
                "version": "1.0.0",
                # Missing 'engine' field
            },
        }

        # Execute and Assert
        with self.assertRaises(ValidationError):
            retrieve_project_manifest(manifest_path)


class TestRetrieveVariablesConfig(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.file_exists")
    def test_checks_file_existence(self, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = False
        variables_config_path = Path("/fake/path/variables.yaml")

        # Execute and Assert
        with self.assertRaises(FileNotFoundError):
            retrieve_variables_config(variables_config_path)

        mock_file_exists.assert_called_once_with(variables_config_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=(
            "schema_version: 1\n"
            "required:\n"
            "  var1: value1\n"
            "required_sensitive:\n"
            "  var2: value2\n"
            "overrides:\n"
            "  var3: value3\n"
            "defaults:\n"
            "  var3: default3"
        ),
    )
    @patch("yaml.safe_load")
    @patch("jupyter_deploy.variables_config.JupyterDeployVariablesConfig")
    def test_open_file_call_safe_load_and_parse(
        self, mock_config_class: Mock, mock_yaml_load: Mock, mock_open_file: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")
        yaml_content = {
            "schema_version": 1,
            "required": {"var1": "value1"},
            "required_sensitive": {"var2": "value2"},
            "overrides": {"var3": "value3"},
            "defaults": {"var3": "default3"},
        }
        mock_yaml_load.return_value = yaml_content
        mock_config = Mock()
        mock_config_class.return_value = mock_config

        # Execute
        result = retrieve_variables_config(variables_config_path)

        # Assert
        mock_open_file.assert_called_once_with(variables_config_path)
        mock_yaml_load.assert_called_once()
        mock_config_class.assert_called_once_with(**yaml_content)
        self.assertEqual(result, mock_config)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open)
    def test_parse_config_versions(self, mock_open_file: Mock, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")

        # Test for schema_version 1
        yaml_content = """
        schema_version: 1
        required:
          var1: value1
        required_sensitive:
          var2: value2
        overrides:
          var3: value3
        defaults:
          var3: default3
        """
        mock_open_file.return_value.read.return_value = yaml_content

        with patch(
            "yaml.safe_load",
            return_value={
                "schema_version": 1,
                "required": {"var1": "value1"},
                "required_sensitive": {"var2": "value2"},
                "overrides": {"var3": "value3"},
                "defaults": {"var3": "default3"},
            },
        ):
            # Execute
            result = retrieve_variables_config(variables_config_path)

            # Assert
            self.assertEqual(result.schema_version, 1)
            self.assertEqual(result.required["var1"], "value1")
            self.assertEqual(result.required_sensitive["var2"], "value2")
            self.assertEqual(result.overrides["var3"], "value3")
            self.assertEqual(result.defaults["var3"], "default3")

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open")
    def test_surfaces_error_when_open_raises_os_error(self, mock_open_file: Mock, mock_file_exists: Mock) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")
        mock_open_file.side_effect = OSError("Permission denied")

        # Execute and Assert
        with self.assertRaises(OSError):
            retrieve_variables_config(variables_config_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open, read_data="invalid: yaml: content:")
    @patch("yaml.safe_load")
    def test_raise_yaml_parse_error_on_invalid_yaml(
        self, mock_yaml_load: Mock, _: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")
        mock_yaml_load.side_effect = ParserError("YAML parsing error")

        # Execute and Assert
        with self.assertRaises(ParserError):
            retrieve_variables_config(variables_config_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open, read_data="- item1\n- item2")
    @patch("yaml.safe_load")
    def test_raise_value_error_when_parsed_content_is_not_a_dict(
        self, mock_yaml_load: Mock, _: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")
        mock_yaml_load.return_value = ["item1", "item2"]  # Not a dict

        # Execute and Assert
        with self.assertRaises(NotADictError):
            retrieve_variables_config(variables_config_path)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("builtins.open", new_callable=mock_open, read_data="schema_version: 1\nwrong_field: missing_required_fields")
    @patch("yaml.safe_load")
    def test_raise_validation_error_when_pydantic_parsing_fails(
        self, mock_yaml_load: Mock, _: Mock, mock_file_exists: Mock
    ) -> None:
        # Setup
        mock_file_exists.return_value = True
        variables_config_path = Path("/fake/path/variables.yaml")
        # Missing required fields in the config
        mock_yaml_load.return_value = {
            "schema_version": 1,
            "wrong_field": "missing_required_fields",
            "required": ["I", "should", "be", "a", "dict"],
        }

        # Execute and Assert
        with self.assertRaises(ValidationError):
            retrieve_variables_config(variables_config_path)



import unittest
from unittest.mock import Mock, patch

from typer.testing import CliRunner

from jupyter_deploy.cli.servers_app import servers_app
from jupyter_deploy.manifest import InvalidServiceError


class TestServersApp(unittest.TestCase):
    def test_help_command(self) -> None:
        self.assertTrue(len(servers_app.info.help or "") > 0, "help should not be empty")

        runner = CliRunner()
        result = runner.invoke(servers_app, ["--help"])

        self.assertEqual(result.exit_code, 0)
        for cmd in ["status", "start", "stop", "restart"]:
            self.assertTrue(result.stdout.index(cmd) > 0, f"missing command: {cmd}")

    def test_no_arg_defaults_to_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(servers_app, [])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(len(result.stdout) > 0)


class TestServerStatusCmd(unittest.TestCase):
    def get_mock_server_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock server handler."""
        mock_get_server_status = Mock()
        mock_get_console = Mock()
        mock_server_handler = Mock()

        mock_server_handler.get_server_status = mock_get_server_status
        mock_server_handler.get_console = mock_get_console

        mock_get_server_status.return_value = "IN_SERVICE"
        mock_get_console.return_value = Mock()

        return mock_server_handler, {
            "get_server_status": mock_get_server_status,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_server_handler_and_call_status(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["status"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_server_handler_class.assert_called_once()
        mock_handler_fns["get_server_status"].assert_called_once()
        mock_handler_fns["get_console"].assert_called_once()

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_uses_handler_console_to_print_status_response(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["status"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("IN_SERVICE" in mock_call[1][0])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, _ = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["status", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_server_handler_get_server_status_raises(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_handler_fns["get_server_status"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["status"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestServerStartCmd(unittest.TestCase):
    def get_mock_server_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock server handler."""
        mock_start_server = Mock()
        mock_get_console = Mock()
        mock_server_handler = Mock()

        mock_server_handler.start_server = mock_start_server
        mock_server_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_server_handler, {
            "start_server": mock_start_server,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_server_handler_and_calls_start(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["start"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_server_handler_class.assert_called_once()
        mock_handler_fns["start_server"].assert_called_once_with("all")
        mock_handler_fns["get_console"].assert_called_once()

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_print_the_valid_services_when_passed_an_invalid_service(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        mock_handler_fns["start_server"].side_effect = InvalidServiceError(
            "Invalid service, use one of ['jupyter', 'traefik']"
        )

        # Set up the console mock
        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["start", "--service", "invalid_service"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("Invalid service" in mock_call[1][0])
        self.assertTrue("red" in mock_call[2]["style"])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, _ = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["start", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_start_raises(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_handler_fns["start_server"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["start"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_service_parameter_passes_service_name_for_start(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["start", "--service", "jupyter"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_handler_fns["start_server"].assert_called_once_with("jupyter")


class TestServerStopCmd(unittest.TestCase):
    def get_mock_server_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock server handler."""
        mock_stop_server = Mock()
        mock_get_console = Mock()
        mock_server_handler = Mock()

        mock_server_handler.stop_server = mock_stop_server
        mock_server_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_server_handler, {
            "stop_server": mock_stop_server,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_server_handler_and_calls_stop(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["stop"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_server_handler_class.assert_called_once()
        mock_handler_fns["stop_server"].assert_called_once_with("all")
        mock_handler_fns["get_console"].assert_called_once()

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_print_the_valid_services_when_passed_an_invalid_service(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        mock_handler_fns["stop_server"].side_effect = InvalidServiceError(
            "Invalid service, use one of ['jupyter', 'traefik']"
        )

        # Set up the console mock
        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["stop", "--service", "invalid_service"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("Invalid service" in mock_call[1][0])
        self.assertTrue("red" in mock_call[2]["style"])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, _ = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["stop", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_stop_raises(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_handler_fns["stop_server"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["stop"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_service_parameter_passes_service_name_for_stop(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["stop", "--service", "jupyter"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_handler_fns["stop_server"].assert_called_once_with("jupyter")


class TestServerRestartCmd(unittest.TestCase):
    def get_mock_server_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock server handler."""
        mock_restart_server = Mock()
        mock_get_console = Mock()
        mock_server_handler = Mock()

        mock_server_handler.restart_server = mock_restart_server
        mock_server_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_server_handler, {
            "restart_server": mock_restart_server,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_server_handler_and_calls_restart(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["restart"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_server_handler_class.assert_called_once()
        mock_handler_fns["restart_server"].assert_called_once_with("all")
        mock_handler_fns["get_console"].assert_called_once()

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_print_the_valid_services_when_passed_an_invalid_service(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        mock_handler_fns["restart_server"].side_effect = InvalidServiceError(
            "Invalid service, use one of ['jupyter', 'traefik']"
        )

        # Set up the console mock
        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["restart", "--service", "invalid_service"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("Invalid service" in mock_call[1][0])
        self.assertTrue("red" in mock_call[2]["style"])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, _ = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["restart", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_restart_raises(self, mock_project_dir: Mock, mock_server_handler_class: Mock) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_handler_fns["restart_server"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["restart"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_service_parameter_passes_service_name_for_restart(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["restart", "--service", "jupyter"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_handler_fns["restart_server"].assert_called_once_with("jupyter")


class TestServerLogsCmd(unittest.TestCase):
    def get_mock_server_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock server handler."""
        mock_get_server_logs = Mock()
        mock_get_console = Mock()
        mock_server_handler = Mock()

        mock_server_handler.get_server_logs = mock_get_server_logs
        mock_server_handler.get_console = mock_get_console

        mock_get_server_logs.return_value = "Sample log output"
        mock_get_console.return_value = Mock()

        return mock_server_handler, {
            "get_server_logs": mock_get_server_logs,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_server_handler_and_calls_get_logs_and_print_results(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_handler_fns["get_server_logs"].return_value = "some-logs", "some-errors"

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["logs"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_server_handler_class.assert_called_once()
        mock_handler_fns["get_server_logs"].assert_called_once_with(service="default", extra=[])
        mock_handler_fns["get_console"].assert_called()
        mock_console.print.assert_called()
        mock_console.rule.assert_called()

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_print_the_valid_services_when_passed_an_invalid_service(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        mock_handler_fns["get_server_logs"].side_effect = InvalidServiceError(
            "Invalid service, use one of ['jupyter', 'traefik']"
        )

        # Set up the console mock
        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["logs", "--service", "invalid_service"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("Invalid service" in mock_call[1][0])
        self.assertTrue("red" in mock_call[2]["style"])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_prints_placeholder_when_no_logs_are_returned(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Set up the logs to be empty
        mock_handler_fns["get_server_logs"].return_value = ""

        # Set up the console mock
        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_handler_fns["get_server_logs"].return_value = "", ""

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["logs"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("no logs were retrieved" in mock_call[1][0])
        self.assertTrue("yellow" in mock_call[2]["style"])

    @patch("jupyter_deploy.handlers.resource.server_handler.ServerHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_raises_when_server_handler_raises(
        self, mock_project_dir: Mock, mock_server_handler_class: Mock
    ) -> None:
        # Setup
        mock_server_handler, mock_handler_fns = self.get_mock_server_handler()
        mock_server_handler_class.return_value = mock_server_handler
        mock_handler_fns["get_server_logs"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(servers_app, ["logs"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)



import hcl2

from jupyter_deploy.engine.terraform import tf_outdefs


def extract_description_from_dot_tf_content(content: str) -> dict[str, str]:
    """Parse the content of a outputs.tf file, return outputs as dict name->description."""
    if not content:
        return {}

    parsed_outputs_dot_tf = hcl2.loads(content)
    parsed_outputs = tf_outdefs.ParsedOutputsDotTf(**parsed_outputs_dot_tf)
    parsed_outputs_definitions = parsed_outputs.output

    result: dict[str, str] = {}

    for idx, parsed_out in enumerate(parsed_outputs_definitions):
        if not isinstance(parsed_out, dict):
            print(f"Warning: parsed output was not a dict at idx: {idx}")
            continue

        out_name = next(iter(parsed_out), None)

        if not out_name or len(parsed_out.keys()) != 1:
            print(f"Warning: parsed output at idx '{idx}' is not a dict of size 1.")
            continue

        out_config = parsed_out[out_name]

        if not isinstance(out_name, str):
            print(f"Warning: parsed output key is not a string: {idx}")
            continue
        if not isinstance(out_config, dict):
            print(f"Warning: parsed output '{out_name}' config is not a dict")
            continue

        description = out_config.get("description", "")
        if not isinstance(description, str):
            print(f"Warning: parsed output '{out_name}' description is not a str")
            description = ""

        result.update({out_name: description})

    return result


def combine_cmd_and_outputs_dot_tf_results(
    output_defs_from_cmd: dict[str, tf_outdefs.TerraformOutputDefinition],
    descriptions_from_file: dict[str, str],
) -> dict[str, tf_outdefs.TerraformOutputDefinition]:
    """Adds to description to the output_defs from cmd, return updated dict."""

    for output_name, output_def_from_cmd in output_defs_from_cmd.items():
        description = descriptions_from_file.get(output_name)

        if description is None:
            print(f"Warning: output '{output_name}' not found in outputs.tf file.")
            output_def_from_cmd.description = ""
            continue

        output_def_from_cmd.description = description

    return output_defs_from_cmd



from pathlib import Path

from jupyter_deploy import cmd_utils, fs_utils
from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.outdefs import TemplateOutputDefinition
from jupyter_deploy.engine.terraform import tf_outdefs, tf_outfiles
from jupyter_deploy.engine.terraform.tf_constants import TF_ENGINE_DIR, TF_OUTPUT_CMD, TF_OUTPUTS_FILENAME
from jupyter_deploy.manifest import JupyterDeployManifest


class TerraformOutputsHandler(EngineOutputsHandler):
    """Terraform-specific implementation of the OutputsHandler."""

    def __init__(self, project_path: Path, project_manifest: JupyterDeployManifest) -> None:
        self.project_path = project_path
        self.project_manifest = project_manifest
        self._full_template_outputs: dict[str, TemplateOutputDefinition] | None = None
        self.engine_dir_path = project_path / TF_ENGINE_DIR

    def get_full_project_outputs(self) -> dict[str, TemplateOutputDefinition]:
        # cache handling to avoid the expensive fs operation necessary
        # to retrieve the output definitions.
        if self._full_template_outputs:
            return self._full_template_outputs

        # execute the terraform output command
        output_cmd = TF_OUTPUT_CMD.copy()
        output_content = cmd_utils.run_cmd_and_capture_output(output_cmd, exec_dir=self.engine_dir_path)
        output_defs_from_cmd = tf_outdefs.parse_output_cmd_result(output_content)

        # read the outputs.tf, retrieve the description
        outputs_dot_tf_path = self.engine_dir_path / TF_OUTPUTS_FILENAME
        output_dot_tf_content = fs_utils.read_short_file(outputs_dot_tf_path)
        description_from_outputs_dot_tf = tf_outfiles.extract_description_from_dot_tf_content(output_dot_tf_content)

        combined_output_defs = tf_outfiles.combine_cmd_and_outputs_dot_tf_results(
            output_defs_from_cmd=output_defs_from_cmd,
            descriptions_from_file=description_from_outputs_dot_tf,
        )

        template_output_defs = {
            output_name: output_def.to_template_definition() for output_name, output_def in combined_output_defs.items()
        }
        self._full_template_outputs = template_output_defs
        return template_output_defs

    def get_output_definition(self, output_name: str) -> TemplateOutputDefinition:
        full_template_outputs = self.get_full_project_outputs()
        output_def = full_template_outputs.get(output_name)

        if not output_def:
            raise ValueError(f"Output name not found: {output_name}")
        return output_def



from enum import Enum

from rich import console as rich_console

from jupyter_deploy.provider.aws.aws_ec2_runner import AwsEc2Runner
from jupyter_deploy.provider.aws.aws_ssm_runner import AwsSsmRunner
from jupyter_deploy.provider.instruction_runner import InstructionRunner
from jupyter_deploy.provider.resolved_argdefs import ResolvedInstructionArgument
from jupyter_deploy.provider.resolved_resultdefs import ResolvedInstructionResult


class AwsService(str, Enum):
    """AWS services mapped to jupyter-deploy instructions."""

    EC2 = "ec2"
    SSM = "ssm"


class AwsApiRunner(InstructionRunner):
    """Wrapper class to provide access to AWS APIs.

    Requires the user to install jupyter-deploy[aws].
    """

    def __init__(self, region_name: str | None) -> None:
        """Instantiate the map of AWS services runner."""
        self.region_name = region_name
        self.service_runners: dict[str, InstructionRunner] = {}

    @staticmethod
    def _get_service_and_sub_instruction_name(instruction_name: str) -> tuple[str, str]:
        """Return a tuple of service-name, instruction-name."""
        # expect aws.<service-name>.<instruction>

        parts = instruction_name.split(".")

        if len(parts) < 2 or not parts[1] or not parts[2]:
            raise ValueError(
                f"Invalid instruction: {instruction_name}; should be of the form aws.service-name.instruction-name"
            )
        return parts[1], ".".join(parts[2:])

    def _get_service_runner(self, service_name: str) -> InstructionRunner:
        service_runner = self.service_runners.get(service_name)

        if service_runner:
            return service_runner

        if service_name == AwsService.SSM:
            service_runner = AwsSsmRunner(region_name=self.region_name)
            self.service_runners[service_name] = service_runner
            return service_runner
        elif service_name == AwsService.EC2:
            service_runner = AwsEc2Runner(region_name=self.region_name)
            self.service_runners[service_name] = service_runner
            return service_runner

        raise NotImplementedError(f"Unrecognized AWS service name: {service_name}")

    def execute_instruction(
        self,
        instruction_name: str,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        service_name, sub_instruction_name = AwsApiRunner._get_service_and_sub_instruction_name(instruction_name)
        service_runner = self._get_service_runner(service_name)
        return service_runner.execute_instruction(
            instruction_name=sub_instruction_name,
            resolved_arguments=resolved_arguments,
            console=console,
        )



from pathlib import Path

from jupyter_deploy import fs_utils
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.infrastructure.enum import AWSInfrastructureType, InfrastructureType
from jupyter_deploy.provider.enum import ProviderType
from jupyter_deploy.template_utils import TEMPLATES


class InitHandler:
    """Base class to manage a project at the target disk location."""

    def __init__(
        self,
        project_dir: str | None,
        engine: EngineType = EngineType.TERRAFORM,
        provider: ProviderType = ProviderType.AWS,
        infrastructure: InfrastructureType = AWSInfrastructureType.EC2,
        template: str = "base",
    ) -> None:
        """Create the project handler."""
        if not project_dir:
            self.project_path = fs_utils.get_default_project_path()
        else:
            self.project_path = Path(project_dir)

        self.abs_project_path = Path.resolve(self.project_path)
        self.engine = engine

        full_template_name: str | None = None
        if provider and infrastructure and template:
            full_template_name = f"{provider.lower()}:{infrastructure.lower()}:{template.lower()}"

        self.source_path = self._find_template_path(full_template_name)

    def _find_template_path(self, template_name: str | None) -> Path:
        """Return the path of the template name.

        The template should be of the form <provider>:<infrastructure>:<template>.

        Raises:
            ValueError: if the template is not found.
        """
        if not template_name:
            raise ValueError("Template name cannot be empty")

        engine_name = self.engine.lower()

        if engine_name not in TEMPLATES:
            available_engines = list(TEMPLATES.keys()) if TEMPLATES else "none available"
            raise ValueError(f"Engine '{engine_name}' is not supported. Available engines: {available_engines}")

        engine_templates = TEMPLATES[engine_name]

        if template_name in engine_templates:
            return engine_templates[template_name]

        raise ValueError(
            f"Template '{template_name}' not found for engine '{engine_name}'. "
            f"Available templates: {list(engine_templates.keys()) if engine_templates else 'none'}"
        )

    def may_export_to_project_path(self) -> bool:
        """Verify that the project output path does not contain any file or sub-directory."""
        if not self.project_path.exists():
            return True
        return fs_utils.is_empty_dir(self.project_path)

    def clear_project_path(self) -> None:
        """Clear the project on disk.

        This method assumes that the user accepted to delete the existing files.
        """
        fs_utils.safe_clean_directory(self.project_path)

    def setup(self) -> None:
        """Copies the files from the source location to the target path of the project."""
        fs_utils.safe_copy_tree(self.source_path, self.project_path)



from collections.abc import Callable

from packaging.version import InvalidVersion, Version
from rich.console import Console

from jupyter_deploy import cmd_utils
from jupyter_deploy.enum import JupyterDeployTool
from jupyter_deploy.manifest import JupyterDeployRequirementV1


def _check_installation(
    tool_name: str, min_version: Version | None = None, installation_url: str | None = None
) -> bool:
    """Shell out to verify tool installation, return True if correct."""

    installed, str_version, error_msg = cmd_utils.check_executable_installation(
        executable_name=tool_name,
    )
    console = Console()

    if not installed:
        console.print(f":x: This operation requires [bold]{tool_name}[/] to be installed in your system.", style="red")
        console.line()

        if error_msg:
            console.print(f"Error: {error_msg}", style="red")
            console.line()

        if installation_url:
            console.print(f"Refer to the installation guide: {installation_url}")
        return False

    if min_version:
        if not str_version:
            console.print(
                f"Current version of [bold]{tool_name}[/] not found, cannot perform minimum version check.",
                style="red",
            )
            return False

        current_version = Version(str_version)
        if current_version >= min_version:
            console.print(f"Valid [bold]{tool_name}[/] installation detected.")
            return True
        else:
            console.print(
                f"This operation requires minimum [bold]{tool_name}[/] version: {min_version}\n"
                f"Found version: {current_version}"
            )
            console.print(f"Upgrade [bold]{tool_name}[/] at least to version: {min_version}.", style="yellow")
            return False

    console.print(f":white_check_mark: Valid [bold]{tool_name}[/] installation detected.")
    return True


def _check_terraform_installation(min_version: Version | None = None) -> bool:
    """Shell out to verify terraform installation, return True if valid."""

    return _check_installation(
        tool_name="terraform",
        min_version=min_version,
        installation_url="https://developer.hashicorp.com/terraform/tutorials/aws-get-started/install-cli",
    )


def _check_aws_cli_installation(min_version: Version | None = None) -> bool:
    """Shell out to verify `aws` install, return True if correct."""
    return _check_installation(
        tool_name="aws",
        min_version=min_version,
        installation_url="https://docs.aws.amazon.com/cli/latest/userguide/getting-started-install.html",
    )


def _check_ssm_plugin_installation(min_version: Version | None = None) -> bool:
    """Shell out to verify `session-manager-plugin` install, return True if correct."""
    return _check_installation(
        tool_name="session-manager-plugin",
        min_version=min_version,
        installation_url="https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html",
    )


def _check_jq_installation(min_version: Version | None = None) -> bool:
    """Shell out to verify `jq` install, return True if correct."""
    return _check_installation(
        tool_name="jq",
        min_version=min_version,
        installation_url="https://jqlang.org/download/",
    )


_TOOL_VERIFICATION_FN_MAP: dict[JupyterDeployTool, Callable[[Version | None], bool]] = {
    JupyterDeployTool.AWS_CLI: _check_aws_cli_installation,
    JupyterDeployTool.AWS_SSM_PLUGIN: _check_ssm_plugin_installation,
    JupyterDeployTool.JQ: _check_jq_installation,
    JupyterDeployTool.TERRAFORM: _check_terraform_installation,
}


def verify_tools_installation(requirements: list[JupyterDeployRequirementV1]) -> bool:
    """Verify all requirements in order, return True if all requirements are satisfied."""
    tool_check_fns: list[tuple[Callable[[Version | None], bool], Version | None]] = []

    for req in requirements:
        try:
            tool = JupyterDeployTool.from_string(req.name)
        except ValueError:
            continue

        try:
            min_version = Version(req.version) if req.version else None
        except InvalidVersion:
            min_version = None

        tool_check_fns.append((_TOOL_VERIFICATION_FN_MAP[tool], min_version))

    verified = True
    for tool_check_fn, tool_min_version in tool_check_fns:
        verified = tool_check_fn(tool_min_version) and verified
    return verified



from enum import Enum


class InstructionArgumentSource(str, Enum):
    """Enum to list the possible sources for an instruction argument."""

    TEMPLATE_OUTPUT = "output"
    CLI_ARGUMENT = "cli"
    INSTRUCTION_RESULT = "result"

    @classmethod
    def from_string(cls, source_str: str) -> "InstructionArgumentSource":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        source_lower = source_str.lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No InstructionArgumentSource found for '{source_str}'")


class ResultSource(str, Enum):
    """Enum to list the possible sources for an result."""

    TEMPLATE_OUTPUT = "output"
    INSTRUCTION_RESULT = "result"

    @classmethod
    def from_string(cls, source_str: str) -> "ResultSource":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        source_lower = source_str.lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No ResultSource found for '{source_str}'")


class ValueSource(str, Enum):
    """Enum to list the possible sources for a declared value."""

    TEMPLATE_OUTPUT = "output"

    @classmethod
    def from_string(cls, source_str: str) -> "ValueSource":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        source_lower = source_str.lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No ValueSource found for '{source_str}'")


class UpdateSource(str, Enum):
    """Enum to list the possible sources for an result."""

    INSTRUCTION_RESULT = "result"

    @classmethod
    def from_string(cls, source_str: str) -> "UpdateSource":
        """Return the enum value, ignoring case.

        Raises:
            ValueError: If no matching enum value is found.
        """
        source_lower = source_str.lower()
        for source in cls:
            if source.value.lower() == source_lower:
                return source
        raise ValueError(f"No UpdateSource found for '{source_str}'")


class TransformType(str, Enum):
    """List of transforms applicable to results value."""

    NO_TRANSFORM = "no-transform"
    COMMA_SEPARATED_STR_TO_LIST_STR = "comma-separated-str-to-list-str"

    @classmethod
    def from_string(cls, target_str: str | None) -> "TransformType":
        """Return the enum value, ignoring case, dashes and underlines.

        Raises:
            ValueError: If no matching enum value is found.
        """
        if not target_str:
            return cls.NO_TRANSFORM

        target_lower = target_str.lower()

        # first try to match lower case only
        for transform in cls:
            if transform.value.lower() == target_lower:
                return transform

        # then try to match without dashes or underlines
        repl_target = target_lower.replace("-", "").replace("_", "")
        for transform in cls:
            repl_transform = transform.value.lower().replace("-", "").replace("_", "")
            if repl_transform == repl_target:
                return transform

        raise ValueError(f"No transform found for '{target_str}'")


class JupyterDeployTool(str, Enum):
    """List of tools verifiable by jupyter deploy."""

    AWS_CLI = "aws-cli"
    AWS_SSM_PLUGIN = "aws-ssm-plugin"
    JQ = "jq"
    TERRAFORM = "terraform"

    @classmethod
    def from_string(cls, target_str: str) -> "JupyterDeployTool":
        """Return the enum value, ignoring case, dashes and underlines.

        Raises:
            ValueError: If no matching enum value is found.
        """
        target_lower = target_str.lower()

        # first try to match lower case only
        for source in cls:
            if source.value.lower() == target_lower:
                return source

        # then try to match without dashes or underlines
        repl_target = target_lower.replace("-", "").replace("_", "")
        for source in cls:
            repl_source = source.value.lower().replace("-", "").replace("_", "")
            if repl_source == repl_target:
                return source

        raise ValueError(f"No tool found for '{target_str}'")



from typing import Any

import hcl2
from pydantic import ValidationError

from jupyter_deploy.engine.terraform import tf_plan, tf_vardefs


def parse_variables_dot_tf_content(content: str) -> dict[str, tf_vardefs.TerraformVariableDefinition]:
    """Parse the content of a variables.tf file, return variables as dict name->var_def."""
    if not content:
        return {}

    parsed_variables_dot_tf = hcl2.loads(content)
    parsed_vars = tf_vardefs.ParsedVariablesDotTf(**parsed_variables_dot_tf)

    parsed_vars_definitions = parsed_vars.variable

    result: dict[str, tf_vardefs.TerraformVariableDefinition] = {}

    for idx, parsed_var in enumerate(parsed_vars_definitions):
        if not isinstance(parsed_var, dict):
            print(f"Warning: parsed variable was not a dict at idx: {idx}")
            continue

        var_name = next(iter(parsed_var), None)

        if not var_name or len(parsed_var.keys()) != 1:
            print(f"Warning: parsed variable at idx '{idx}' is not a dict of size 1.")
            continue

        var_config = parsed_var[var_name]

        if not isinstance(var_name, str):
            print(f"Warning: parsed variable key is not a string: {idx}")
            continue
        if not isinstance(var_config, dict):
            print(f"Warning: parsed variable '{var_name}' config is not a dict")
            continue

        tf_type = var_config.pop("type")
        var_config["tf_type"] = tf_type
        var_config["variable_name"] = var_name

        try:
            var_def = tf_vardefs.create_tf_variable_definition(var_config)
            result.update({var_name: var_def})
        except ValidationError as e:
            print(f"Error parsing variable definition for '{var_name}': {e.errors()}")
            continue
        except RuntimeError as e:
            print(f"Skipping unhandled variable type. {e}")
            continue

    return result


def parse_dot_tfvars_content_and_add_defaults(
    content: str,
    variable_defs: dict[str, tf_vardefs.TerraformVariableDefinition],
) -> None:
    """Parse the content of a .tfvars, add as default to the previously parsed variables definition.

    Modify the 'variable_defs' entries in place.

    This method skips the value of all sensitive variables, as defined in the 'variable_defs'
    parameter.
    """
    parsed_vars_name_default_value_map = hcl2.loads(content)

    for var_name, var_default in parsed_vars_name_default_value_map.items():
        varname = var_name if isinstance(var_name, str) else None

        if not varname:
            print(f"Warning: variable name '{varname}' in .tfvars should be a string; skippin.")
            continue

        if varname not in variable_defs:
            print(f"Warning: variable '{var_name}' in .tfvars file not found in 'variables.tf'; skipping.")
            continue

        var_def = variable_defs[varname]

        if var_def.sensitive:
            print(f"Warning: found value in .tfvars file for sensitive variable '{varname}'; ignoring.")
            continue

        try:
            updated_var_def = tf_vardefs.create_tf_variable_definition(
                {**var_def.model_dump(), "default": var_default, "has_default": True}
            )
            variable_defs[varname] = updated_var_def
        except ValidationError as e:
            print(f"Warning: invalid default in .tfvars file for '{varname}', ignoring: {e.errors()}")
            continue


def parse_and_update_dot_tfvars_content(
    content: str,
    varvalues: dict[str, Any],
) -> list[str]:
    """Load a dot tfvars file, add the values to the dict, return the lines of the new content."""
    if not content:
        saved_values: dict[str, Any] = {}
    else:
        saved_values = hcl2.loads(content)
    saved_values.update(varvalues)
    return tf_plan.format_values_for_dot_tfvars(saved_values)


def parse_and_remove_overridden_variables_from_content(
    content: str,
    varnames_to_remove: list[str],
) -> list[str]:
    """Load a dot tfvars file, remove keys from the dict, return the lines of the new content."""
    if not content:
        saved_values: dict[str, Any] = {}
    else:
        saved_values = hcl2.loads(content)
    for varname in varnames_to_remove:
        if varname in saved_values:
            del saved_values[varname]
    return tf_plan.format_values_for_dot_tfvars(saved_values)
